import time

import pytest

from flows_e2e_tests.config import settings
from flows_e2e_tests.utils import FlowResponse, RunResponse, flows_client


@pytest.mark.slow
def test_run_with_consent_required_can_resume_to_completion(
    flow_for_tests: FlowResponse, run_for_tests: RunResponse
):
    # Wait for the Run to become inactive
    for _ in range(settings.max_wait_time):
        run_status_resp = flows_client.flow_action_status(
            flow_id=flow_for_tests.id,
            flow_scope=flow_for_tests.globus_auth_scope,
            flow_action_id=run_for_tests.run_id,
        )
        if (
            run_status_resp.data["status"] == "INACTIVE"
            and run_status_resp.data["details"]["code"] == "ConsentRequired"
        ):
            consent_scope = run_status_resp["details"]["required_scope"]
            break
        time.sleep(1)
    else:
        # We only execute this if the Action never reaches its INACTIVE state by
        # the time max_wait_time has elapsed
        assert False, "Action never reached INACTIVE state due to ConsentRequired"

    # Attempt to resume the Run with updated consents
    resume_resp = flows_client.flow_action_resume(
        flow_id=flow_for_tests.id,
        flow_scope=consent_scope,
        flow_action_id=run_for_tests.run_id,
    )
    assert resume_resp.http_status == 200
    assert resume_resp.data["status"] == "ACTIVE"

    # Wait for the Run to complete
    for _ in range(settings.max_wait_time):
        run_status_resp = flows_client.flow_action_status(
            flow_id=flow_for_tests.id,
            flow_scope=flow_for_tests.globus_auth_scope,
            flow_action_id=run_for_tests.run_id,
        )
        if run_status_resp.data["status"] == "SUCCEEDED":
            break
        time.sleep(1)
    else:
        # We only execute this if the Action never reaches its INACTIVE state by
        # the time max_wait_time has elapsed
        assert False, "Action never reached ACTIVE state after resuming"

    assertions = (
        run_status_resp.data["status"] == "SUCCEEDED"
        and run_status_resp.data["completion_time"] is not None
        and run_status_resp.data["details"]["code"] == "FlowSucceeded"
    )
    assert assertions, run_status_resp.data



import argparse
import uuid

import pytest
import structlog

from flows_e2e_tests.config import settings
from flows_e2e_tests.utils import flows_client

logger = structlog.get_logger(__name__)


def run_tests():
    parser = argparse.ArgumentParser(description="Run Flows E2E tests")
    parser.add_argument("--no-slow", action="store_true", help="Don't run slow tests")
    parser.add_argument(
        "--debug", action="store_true", help="Display discovered test settings"
    )
    args = parser.parse_args()

    if args.debug:
        print(f"Loaded settings={settings.censored}")
    elif args.no_slow:
        pytest.main(["-m not slow"])
    else:
        pytest.main()


def delete_flow():
    parser = argparse.ArgumentParser(description="Delete one or more Flows")
    parser.add_argument(
        "flow_ids",
        metavar="FLOW_ID",
        type=uuid.UUID,
        nargs="+",
    )
    args = parser.parse_args()
    for f_id in args.flow_ids:
        flows_client.delete_flow(f_id)

    logger.info(f"Deleted Flow(s) with ID(s) {args.flow_ids}")



from flows_e2e_tests.utils import FlowResponse, flows_client


def test_deployed_flow_can_be_retrieved(flow_for_tests: FlowResponse):
    get_resp = flows_client.get_flow(flow_for_tests.id)
    assert get_resp.http_status == 200

    for key in [
        "id",
        "definition",
        "globus_auth_username",
        "input_schema",
    ]:
        assert get_resp.data[key] == getattr(flow_for_tests, key), get_resp.data


def test_get_flow_conforms_to_schema(flow_for_tests: FlowResponse):
    get_resp = flows_client.get_flow(flow_for_tests.id)
    assert get_resp.http_status == 200
    FlowResponse(**get_resp.data)



import typing as t

import pytest

from flows_e2e_tests.utils import FlowResponse, RunResponse, flows_client

flow_definition = {
    "StartAt": "PassState",
    "States": {
        "PassState": {
            "Type": "Pass",
            "ResultPath": "$.FlowResult",
            "End": True,
        }
    },
}

flow_input: t.Dict[str, t.Any] = {}

metadata = {
    "input_schema": {"additionalProperties": False},
    # "flow_administrators": [],
    # "flow_starters": [],
    # "flow_viewers": [],
    "keywords": [],
}


@pytest.fixture(scope="module")
def flow_with_runs(request):
    deploy_resp = flows_client.deploy_flow(
        title=f"Integration Test Flow for {request.module.__name__}",
        flow_definition=flow_definition,
        **metadata,
    )
    assert deploy_resp.http_status == 201
    flow = FlowResponse(**deploy_resp.data)

    runs: t.List[RunResponse] = []
    for _ in range(1):
        run_resp = flows_client.run_flow(
            flow_id=flow.id,
            flow_scope=flow.globus_auth_scope,
            flow_input=flow_input,
            label=f"Integration Test Run for {request.module.__name__}",
        )
        assert run_resp.http_status == 201
        run = RunResponse(**run_resp.data)
        runs.append(run)

    delete_resp = flows_client.delete_flow(flow.id)
    assert delete_resp.http_status == 200
    yield flow, runs



from flows_e2e_tests.utils import FlowResponse, auth_client

from .conftest import metadata


def test_deployed_flow_has_correct_rbac(flow_for_tests: FlowResponse):
    assert flow_for_tests.user_role == "flow_owner"
    assert (
        flow_for_tests.flow_owner == f"urn:globus:auth:identity:{auth_client.client_id}"
    )


def test_deployed_flow_has_correct_metadata(flow_for_tests: FlowResponse):
    for key, value in metadata.items():
        assert getattr(flow_for_tests, key) == value



import datetime
import typing as t

from pydantic import BaseModel, Field, validator

uuid_regex = (
    "([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})"
)
auth_id_urn_regex = f"^urn:globus:auth:identity:{uuid_regex}$"
principal_urn_regex = f"^urn:globus:(auth:identity|groups:id):{uuid_regex}$"


class FlowResponse(BaseModel):
    id: str = Field(regex=uuid_regex)
    api_version: t.Literal["1.0"]
    created_at: datetime.datetime
    definition: t.Dict[str, t.Any]
    description: str
    flow_administrators: t.List[str]
    flow_owner: str = Field(regex=auth_id_urn_regex)
    flow_starters: t.List[str]
    flow_url: str
    flow_viewers: t.List[str]
    globus_auth_scope: str
    globus_auth_username: str
    # If the Flow doesn't have an input schema defined it will be removed from
    # the output, so default it here
    input_schema: t.Dict[str, t.Any] = {}
    keywords: t.List[str]
    log_supported: bool
    principal_urn: str = Field(regex=auth_id_urn_regex)
    subtitle: str
    synchronous: bool
    title: str
    types: t.List[str]
    updated_at: datetime.datetime
    user_role: t.Literal[
        "flow_owner", "flow_starter", "flow_administrator", "flow_viewer"
    ]

    @validator("globus_auth_scope")
    def validate_scope(cls, v, values, **kwargs):
        flow_id = values["id"]
        if (
            v
            != f"https://auth.globus.org/scopes/{flow_id}/flow_{flow_id.replace('-', '_')}_user"
        ):
            raise ValueError(f"Invalid Globus Auth scope {v}")
        return v

    @validator("globus_auth_username")
    def validate_username(cls, v, values, **kwargs):
        flow_id = values["id"]
        if v != f"{flow_id}@clients.auth.globus.org":
            raise ValueError(f"Invalid Globus Auth username {v}")
        return v

    @validator("flow_url")
    def validate_url(cls, v, values, **kwargs):
        flow_id = values["id"]
        if (
            not v.endswith(f"flows.automate.globus.org/flows/{flow_id}")
            and not v.endswith(f"flows.automate.globus.org/{flow_id}")
            and not v.endswith(f"flows.globus.org/flows/{flow_id}")
            and not v.endswith(f"flows.globus.org/{flow_id}")
        ):
            raise ValueError(f"Invalid Flow URL {v}")
        return v


class RunResponse(BaseModel):
    run_id: str = Field(regex=uuid_regex)
    completion_time: t.Optional[datetime.datetime]
    details: t.Dict[str, t.Any]
    display_status: str
    flow_last_updated: t.Optional[datetime.datetime]
    flow_deleted_at: t.Optional[datetime.datetime] = None
    run_managers: t.List[str] = Field(regex=principal_urn_regex)
    run_monitors: t.List[str] = Field(regex=principal_urn_regex)
    run_owner: str = Field(regex=auth_id_urn_regex)
    start_time: datetime.datetime
    status: t.Literal["ACTIVE", "INACTIVE", "FAILED", "SUCCEEDED"]
    user_role: t.Literal["run_owner", "run_monitor", "run_manager"]

    @validator("completion_time", pre=True)
    def validate_completion_time(cls, v):
        """
        The API returns 'None' as a completion time when the Run is incomplete.
        Help pydantic parse that as None.
        """
        if v == str(None):
            return None
        return v



import typing as t

import structlog
from globus_sdk import ConfidentialAppAuthClient
from globus_sdk.auth.token_response import OAuthTokenResponse
from globus_sdk.authorizers import RefreshTokenAuthorizer

from flows_e2e_tests.config import settings

logger = structlog.get_logger(__name__)

auth_client = ConfidentialAppAuthClient(
    settings.globus_auth_client_id, settings.globus_auth_client_secret
)
logger.debug(f"Initialized Auth Client at {auth_client.base_url}")


def authorizers_by_scope(scopes: t.List[str]) -> t.Dict[str, RefreshTokenAuthorizer]:
    authzs: t.Dict[str, RefreshTokenAuthorizer] = {}
    response: OAuthTokenResponse = auth_client.oauth2_client_credentials_tokens(
        " ".join(scopes)
    )
    for s in scopes:
        base_scope = _get_base_scope(s)
        tokens = response.by_scopes[base_scope]
        auth = RefreshTokenAuthorizer(
            auth_client=auth_client,
            refresh_token=tokens["refresh_token"],
            access_token=tokens["access_token"],
            expires_at=tokens["expires_at_seconds"],
        )
        authzs[s] = authzs[base_scope] = auth
    return authzs


def _get_base_scope(scope: str):
    if "[" in scope:
        return scope.split("[")[0]
    return scope


def authorizer_for_scope(scope: str) -> RefreshTokenAuthorizer:
    return authorizers_by_scope([scope]).get(scope)


def get_token_for_scope(scope: str):
    ...



import pytest
from globus_sdk import GlobusAPIError

from flows_e2e_tests.utils import FlowResponse, flows_client


def test_flow_delete_succeeds(flow_for_one_test: FlowResponse):
    del_resp = flows_client.delete_flow(flow_for_one_test.id)

    assert del_resp.http_status == 200
    with pytest.raises(GlobusAPIError) as err:
        flows_client.get_flow(flow_for_one_test.id)
    assert err.value.http_status == 404



from flows_e2e_tests.utils import FlowResponse, auth_client, flows_client

from .conftest import flow_input


def test_deployed_flow_can_be_run_by_creator(flow_for_tests: FlowResponse):
    run_resp = flows_client.run_flow(
        flow_for_tests.id,
        flow_scope=flow_for_tests.globus_auth_scope,
        flow_input=flow_input,
    )
    assert run_resp.http_status == 201
    assert (
        run_resp.data["created_by"]
        == f"urn:globus:auth:identity:{auth_client.client_id}"
    )



import os
from pathlib import Path

import structlog
from dynaconf import Validator

from .custom import CensoredDynaconf

logger = structlog.get_logger(__name__)

PROJECT_DIR = Path(__file__).resolve().parent
EXECUTION_DIR = Path.cwd()
logger.debug(f"Reading base settings at {PROJECT_DIR}")
logger.debug(f"Searching for overrides at {EXECUTION_DIR}/.env")

ENV_SWITCHER = "FLOWS_TEST_ENVIRONMENT"
ENV_VALUE = os.getenv(ENV_SWITCHER, "production")
if ENV_SWITCHER != "GLOBUS_SDK_ENVIRONMENT":
    os.environ["GLOBUS_SDK_ENVIRONMENT"] = ENV_VALUE
logger.debug(f"Using {ENV_SWITCHER}={ENV_VALUE} to determine settings")


settings = CensoredDynaconf(
    default_env="production",
    envvar_prefix="TEST",
    settings_files=[
        PROJECT_DIR / "settings.toml",
        PROJECT_DIR / ".secrets.toml",
    ],
    dotenv_path=EXECUTION_DIR,
    load_dotenv=True,
    env_switcher=ENV_SWITCHER,
    ignore_unknown_envvars=True,
    environments=True,
    validators=[
        Validator("GLOBUS_AUTH_CLIENT_ID", "GLOBUS_AUTH_CLIENT_SECRET", must_exist=True)
    ],
)



import structlog
from globus_automate_client import create_flows_client
from globus_automate_client.flows_client import MANAGE_FLOWS_SCOPE
from globus_sdk import GlobusAPIError
from globus_sdk.authorizers import RefreshTokenAuthorizer

from flows_e2e_tests.config import settings

from .auth import authorizer_for_scope

logger = structlog.get_logger(__name__)


def authorizer_retriever(*args, flow_scope: str, **kwargs) -> RefreshTokenAuthorizer:
    return authorizer_for_scope(flow_scope)


try:
    flows_client = create_flows_client(
        base_url=None,
        authorizer=authorizer_retriever(flow_scope=MANAGE_FLOWS_SCOPE),
        authorizer_callback=authorizer_retriever,
        http_timeout=10,
    )
except GlobusAPIError as err:
    logger.exception("Unable to create a FlowsClient", settings=settings.as_dict())
    raise SystemExit from err
else:
    logger.debug(f"Initialized Flows Client at {flows_client.base_url}")



import typing as t

import pytest

from flows_e2e_tests.utils import FlowResponse, flows_client

flow_definition = {
    "StartAt": "PassState",
    "States": {
        "PassState": {
            "Type": "Pass",
            "ResultPath": "$.FlowResult",
            "End": True,
        }
    },
}

flow_input: t.Dict[str, t.Any] = {}

metadata = {
    "input_schema": {"additionalProperties": False},
    # "flow_administrators": [],
    # "flow_starters": [],
    # "flow_viewers": [],
    "keywords": [],
}


@pytest.fixture(scope="function")
def flow_for_one_test(request):
    deploy_resp = flows_client.deploy_flow(
        title=f"Integration Test Flow for {request.module.__name__}",
        flow_definition=flow_definition,
        **metadata,
    )
    assert deploy_resp.http_status == 201
    yield FlowResponse(**deploy_resp.data)


@pytest.fixture(scope="module")
def flow_for_tests(request):
    deploy_resp = flows_client.deploy_flow(
        title=f"Integration Test Flow for {request.module.__name__}",
        flow_definition=flow_definition,
        **metadata,
    )
    assert deploy_resp.http_status == 201
    yield (flow := FlowResponse(**deploy_resp.data))

    delete_resp = flows_client.delete_flow(flow.id)
    assert delete_resp.http_status == 200



from flows_e2e_tests.utils import FlowResponse, flows_client


def test_deployed_flow_can_get_updated_by_creator(flow_for_tests: FlowResponse):
    updated_def = flow_for_tests.definition
    updated_def["Comment"] = "This comment was updated"

    updates = {
        "flow_definition": updated_def,
        "title": "This title was updated",
        "subtitle": "This subtitle was updated",
        "description": "This description was updated",
    }
    update_resp = flows_client.update_flow(flow_for_tests.id, **updates)
    assert update_resp.http_status == 200
    for key, value in updates.items():
        # The SDK kwarg is flow_definition but the API's return value is
        # definition, standardize the key so that we only look for definition.
        if key == "flow_definition":
            key = key.strip("flow_")
        assert update_resp.data[key] == value



from dynaconf import Dynaconf


class CensoredDynaconf(Dynaconf):
    dict_fields = ["globus_auth_client_id", "max_wait_time"]
    censored_fields = ["globus_auth_client_secret"]
    censored_field_length = 10

    @property
    def censored(self):
        settings = {}
        for f in self.dict_fields:
            settings[f] = getattr(self, f)

        for f in self.censored_fields:
            value = self._censor_value(getattr(self, f))
            settings[f] = value
        return settings

    def _censor_value(self, value: str):
        value = "*" * 20 + value[-5:]
        return value[-self.censored_field_length :]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flows_e2e_tests',
 'flows_e2e_tests.config',
 'flows_e2e_tests.scenarios.consent_required',
 'flows_e2e_tests.scenarios.flow_lifecycle',
 'flows_e2e_tests.scenarios.soft_delete',
 'flows_e2e_tests.utils']

package_data = \
{'': ['*']}

install_requires = \
['dynaconf>=3.1.4,<4.0.0',
 'globus-automate-client>=0.11.5,<0.12.0',
 'pydantic>=1.8.2,<2.0.0',
 'pytest-xdist>=2.3.0,<3.0.0',
 'pytest>=6.2.4,<7.0.0',
 'structlog>=21.1.0,<22.0.0']

entry_points = \
{'console_scripts': ['globus-flows-e2e-tests = '
                     'flows_e2e_tests.scripts:run_tests',
                     'globus-flows-rm-flow = '
                     'flows_e2e_tests.scripts:delete_flow']}

setup_kwargs = {
    'name': 'flows-e2e-tests',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Uriel Mandujano',
    'author_email': 'uriel@globus.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from flows_e2e_tests.config import settings


def action_provider_url_for_environment(s: str) -> str:
    protocol, fqdn = s.split("://")
    if settings.current_env != "production" and not fqdn.startswith(
        settings.current_env
    ):
        return f"{protocol}://{settings.current_env}.{fqdn}"
    return s



from .auth import auth_client, authorizer_for_scope, authorizers_by_scope
from .flows import flows_client
from .models import FlowResponse, RunResponse
from .providers import action_provider_url_for_environment

all = [
    "action_provider_url_for_environment",
    "auth_client",
    "authorizer_for_scope",
    "authorizers_by_scope",
    "flows_client",
    "FlowResponse",
    "RunResponse",
]



import typing as t

import pytest
from globus_sdk.exc import GlobusAPIError

from flows_e2e_tests.utils import FlowResponse, RunResponse, flows_client

from .conftest import flow_input


def test_runs_persist_after_flow_delete(
    flow_with_runs: t.Tuple[FlowResponse, t.List[RunResponse]],
):
    flow, runs = flow_with_runs
    response = flows_client.enumerate_actions(filters={"filter_flow_id": flow.id})

    run_ids = [r.run_id for r in runs]
    for r_resp in response.data["runs"]:
        r = RunResponse(**r_resp)
        assert r.run_id in run_ids
        assert r.flow_deleted_at is not None


def test_run_operations_work_after_delete(
    flow_with_runs: t.Tuple[FlowResponse, t.List[RunResponse]]
):
    flow, runs = flow_with_runs

    response = flows_client.flow_action_status(
        flow.id, flow.globus_auth_scope, runs[0].run_id
    )
    assert response.http_status == 200, response.data

    response = flows_client.flow_action_log(
        flow.id, flow.globus_auth_scope, runs[0].run_id
    )
    assert response.http_status == 200, response.data

    response = flows_client.flow_action_resume(
        flow.id, flow.globus_auth_scope, runs[0].run_id
    )
    assert response.http_status == 200, response.data

    response = flows_client.flow_action_cancel(
        flow.id, flow.globus_auth_scope, runs[0].run_id
    )
    assert response.http_status == 202, response.data

    response = flows_client.flow_action_release(
        flow.id, flow.globus_auth_scope, runs[0].run_id
    )
    assert response.http_status == 200, response.data


def test_flow_cannot_be_retrieved_after_delete(
    flow_with_runs: t.Tuple[FlowResponse, t.List[RunResponse]]
):
    flow, _ = flow_with_runs
    with pytest.raises(GlobusAPIError) as err:
        flows_client.get_flow(flow.id)
    assert err.value.http_status == 404


def test_flow_cannot_be_run_after_delete(
    flow_with_runs: t.Tuple[FlowResponse, t.List[RunResponse]]
):
    flow, _ = flow_with_runs
    with pytest.raises(GlobusAPIError) as err:
        flows_client.run_flow(
            flow.id, flow_scope=flow.globus_auth_scope, flow_input=flow_input
        )
    assert err.value.http_status == 404


def test_flow_cannot_be_removed_after_delete(
    flow_with_runs: t.Tuple[FlowResponse, t.List[RunResponse]]
):
    flow, _ = flow_with_runs
    with pytest.raises(GlobusAPIError) as err:
        flows_client.delete_flow(flow.id)
    assert err.value.http_status == 404


def test_flow_is_not_listed_after_delete(
    flow_with_runs: t.Tuple[FlowResponse, t.List[RunResponse]]
):
    flow, _ = flow_with_runs
    response = flows_client.list_flows()
    for f in response.data["flows"]:
        f = FlowResponse(**f)
        assert f.id != flow.id


def test_flow_cannot_be_updated_after_delete(
    flow_with_runs: t.Tuple[FlowResponse, t.List[RunResponse]]
):
    flow, _ = flow_with_runs
    with pytest.raises(GlobusAPIError) as err:
        flows_client.update_flow(flow.id, title="Updated Flow Title")
    assert err.value.http_status == 404



import typing as t

import pytest

from flows_e2e_tests.utils import (
    FlowResponse,
    RunResponse,
    action_provider_url_for_environment,
    flows_client,
)

flow = {
    "StartAt": "HelloState",
    "States": {
        "HelloState": {
            "Type": "Action",
            "ActionUrl": action_provider_url_for_environment(
                "https://actions.automate.globus.org/hello_world"
            ),
            "Parameters": {
                "required_dependent_scope": "urn:globus:auth:scope:groups.api.globus.org:all"
            },
            "ResultPath": "$.FlowResult",
            "End": True,
        }
    },
}

flow_input: t.Dict[str, t.Any] = {}

metadata = {
    "input_schema": {"additionalProperties": False},
    # "flow_administrators": [],
    # "flow_starters": [],
    # "flow_viewers": [],
    "keywords": [],
}


@pytest.fixture(scope="module")
def flow_for_tests(request):
    deploy_resp = flows_client.deploy_flow(
        title=f"Integration Test Flow for {request.module.__name__}",
        flow_definition=flow,
        **metadata,
    )
    assert deploy_resp.http_status == 201
    yield FlowResponse(**deploy_resp.data)

    delete_resp = flows_client.delete_flow(deploy_resp.data["id"])
    assert delete_resp.http_status == 200


@pytest.fixture(scope="module")
def run_for_tests(request, flow_for_tests: FlowResponse):
    run_resp = flows_client.run_flow(
        flow_id=flow_for_tests.id,
        flow_scope=flow_for_tests.globus_auth_scope,
        flow_input=flow_input,
        label=f"Integration Test Run for {request.module.__name__}",
    )
    assert run_resp.http_status == 201
    yield RunResponse(**run_resp.data)
